unit main;

////////////////////////////////////////////////////////////////////////////////
// main.pas
// Reseni ulohy celostatniho kola Souteze v programovani 2015
// soubor k formulari F_Main, zde je vpodstate vse
// (c) Jan Horacek
// Janske Lazne 2015
// me@apophis.cz
// 27.06.2015
// v1.0
////////////////////////////////////////////////////////////////////////////////

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, IdHttp, DXDraws, StdCtrls, AppEvnts, StrUtils, Generics.Collections,
  ExtCtrls;

const
  _MAP_SIZE = 10;

type
  TMapPolickoTyp = (prazdne = 0, zed = 1, poklad = 2);
  TMove = (doleva = 0, doprava = 1, dolu = 2, nahoru = 3);
  TUIStatus = (off = 0, mam_cestu_k_pokladu = 1, vracim_se_zpet = 2);

  TMapPolicko = record
    typ:TMapPolickoTyp;
    hodnota_pokladu:Integer;
  end;

  TMap = record
    mapa:array[0.._MAP_SIZE-1, 0.._MAP_SIZE-1] of TMapPolicko;
  end;

  TLovec = record
    pos:TPoint;
    nese_hodnotu:Integer;
    prineseno:Integer;
    base:TPoint;
  end;

  TPath = TList<TMove>;
  TPointList = TList<TPoint>;

  TUI = record
   status:TUIStatus;
   cmds:TPath;
  end;

  TSolution = record
   poklady:TPointList;
   vydelek:Integer;
  end;

  TSolTree = record
    poklady:TList<TPoint>;
    zisk:Integer;
  end;

  TF_Main = class(TForm)
    MM_Main: TMainMenu;
    MI_Options: TMenuItem;
    MI_NewGame: TMenuItem;
    DXDraw1: TDXDraw;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    L_Nese: TLabel;
    L_Odneseno: TLabel;
    L_tahu_celkem: TLabel;
    L_Tahu_zbyva: TLabel;
    ApplicationEvents1: TApplicationEvents;
    T_UI: TTimer;
    GB_Pohyb: TGroupBox;
    B_Up: TButton;
    B_Right: TButton;
    B_Down: TButton;
    B_Left: TButton;
    CHB_Keyboard: TCheckBox;
    GB_UI: TGroupBox;
    B_UI_Step: TButton;
    CHB_AutoUI: TCheckBox;
    Button1: TButton;
    CHB_DeepMode: TCheckBox;
    procedure MI_OptionsClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure MI_NewGameClick(Sender: TObject);
    procedure B_RightClick(Sender: TObject);
    procedure B_LeftClick(Sender: TObject);
    procedure B_DownClick(Sender: TObject);
    procedure B_UpClick(Sender: TObject);
    procedure CHB_KeyboardClick(Sender: TObject);
    procedure ApplicationEvents1Message(var Msg: tagMSG; var Handled: Boolean);
    procedure B_UI_StepClick(Sender: TObject);
    procedure T_UITimer(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    parsed:TStrings;
    parsed2:TStrings;

  public
    map:TMap;
    game_id:Integer;
    zbyva_tahu:Integer;
    me:TLovec;
    other:TLovec;
    other_enabled:boolean;
    UI:TUI;
    best_solution:TSolTree;

    procedure NewGame();
    procedure ShowMap();
    procedure ParseResponse(text:string);
    procedure ShowInfo();
    procedure SendMove(move:TMove);
    procedure CheckMvBtnEnabled();

    function GetPath(start,fin:TPoint; ignore_poklad:boolean = false):TPath;
    function GetPathLen(start,fin:TPoint; ignore_poklad:boolean = false):Integer;
    function VyberNejlepsiPoklad():TPoint;
    function VyberDalsiPoklad(aktualni_delka_cesty:Integer):TSolTree;    // vraci zisk v rekurzivni vetvi
    procedure VyberNejlepsiReseni();
    procedure DoUI();
  end;

var
  F_Main: TF_Main;

implementation

uses Options;

{$R *.dfm}

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.ApplicationEvents1Message(var Msg: tagMSG;
  var Handled: Boolean);
begin
 if (msg.message = WM_KeyDown) then                          //pokud je stisknuta klavesa
  begin
   if ((CHB_Keyboard.Checked) and (Self.game_id > 0)) then
    begin
     case  msg.wParam of                                     //case moznosti stisknutych klaves
       VK_LEFT  : if ((Self.me.pos.X > 0) and (Self.map.mapa[Self.me.pos.X-1][Self.me.pos.Y].typ <> zed)) then Self.SendMove(doleva);
       VK_RIGHT : if ((Self.me.pos.X < _MAP_SIZE-1) and (Self.map.mapa[Self.me.pos.X+1][Self.me.pos.Y].typ <> zed)) then Self.SendMove(doprava);
       VK_UP    : if ((Self.me.pos.Y > 0) and (Self.map.mapa[Self.me.pos.X][Self.me.pos.Y-1].typ <> zed)) then Self.SendMove(nahoru);
       VK_DOWN  : if ((Self.me.pos.Y < _MAP_SIZE-1) and (Self.map.mapa[Self.me.pos.X][Self.me.pos.Y+1].typ <> zed)) then Self.SendMove(dolu);
     end;
    end;
  end;
end;//procedure

procedure TF_Main.B_UI_StepClick(Sender: TObject);
begin
 Self.DoUI();
end;

procedure TF_Main.Button1Click(Sender: TObject);
begin
 Self.VyberNejlepsiReseni();
end;

procedure TF_Main.B_DownClick(Sender: TObject);
begin
 Self.SendMove(dolu);
end;

procedure TF_Main.B_LeftClick(Sender: TObject);
begin
 Self.SendMove(doleva);
end;

procedure TF_Main.B_RightClick(Sender: TObject);
begin
 Self.SendMove(doprava);
end;

procedure TF_Main.B_UpClick(Sender: TObject);
begin
 Self.SendMove(nahoru);
end;

procedure TF_Main.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 F_Options.SaveConfig(_CONFIG_FILE);
end;

procedure TF_Main.FormCreate(Sender: TObject);
begin
 Self.game_id := 0;
 Self.parsed := TStringList.Create();
 Self.parsed2 := TStringList.Create();
end;

procedure TF_Main.FormDestroy(Sender: TObject);
begin
 Self.parsed.Free();
 Self.parsed2.Free();
end;

procedure TF_Main.MI_NewGameClick(Sender: TObject);
begin
 Self.UI.status := TUIStatus.off;
 FreeAndNil(Self.UI.cmds);
 Self.NewGame();
end;

procedure TF_Main.MI_OptionsClick(Sender: TObject);
begin
 F_Options.ShowModal();
end;

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.NewGame();
var IdHTTP: TIdHTTP;
begin
 IdHTTP:= TIdHTTP.Create(nil);
 try
   Self.ParseResponse(IdHTTP.Get('http://'+F_Options.E_hostname.Text+':'+IntToStr(F_Options.SE_Port.Value)+'/start'));
 finally
   IdHTTP.Free;
 end;

 Self.L_tahu_celkem.Caption := IntToStr(Self.zbyva_tahu);
end;//procedure

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.ShowMap();
const
  _SIRKA_POLICKA = 20;      // sirka policka v pixelech
var
  x, y:Integer;
begin
 try
   Self.DXDraw1.Surface.Canvas.Pen.Color := clBlack;

   for y := 0 to _MAP_SIZE-1 do
    begin
     for x := 0 to _MAP_SIZE - 1 do
      begin
       if ((x = Self.me.pos.X) and (y = Self.me.pos.Y)) then
         Self.DXDraw1.Surface.Canvas.Brush.Color := clLime
       else if ((x = Self.me.base.X) and (y = Self.me.base.Y)) then
         Self.DXDraw1.Surface.Canvas.Brush.Color := clBlue
       else if ((x = Self.other.pos.X) and (y = Self.other.pos.Y)) then
         Self.DXDraw1.Surface.Canvas.Brush.Color := clGreen
       else begin


         case (Self.map.mapa[x][y].typ) of
          TMapPolickoTyp.prazdne  : Self.DXDraw1.Surface.Canvas.Brush.Color := clWhite;
          TMapPolickoTyp.zed      : Self.DXDraw1.Surface.Canvas.Brush.Color := clRed;
          TMapPolickoTyp.poklad   : Self.DXDraw1.Surface.Canvas.Brush.Color := clYellow;
         end;//case
        end;// else me pos

       Self.DXDraw1.Surface.Canvas.Rectangle(x*_SIRKA_POLICKA, y*_SIRKA_POLICKA, (x+1)*_SIRKA_POLICKA, (y+1)*_SIRKA_POLICKA);
       if (Self.map.mapa[x][y].typ = poklad) then Self.DXDraw1.Surface.Canvas.TextOut(x*_SIRKA_POLICKA+5, y*_SIRKA_POLICKA+5, IntToStr(Self.map.mapa[x][y].hodnota_pokladu));
      end;//for x
    end;//for y


 finally
   Self.DXDraw1.Surface.Canvas.Release();
   Self.DXDraw1.Flip;
 end;

end;

procedure TF_Main.T_UITimer(Sender: TObject);
begin
 Self.T_UI.Enabled := false;
 Self.DoUI();
end;//procedure

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.ParseResponse(text:string);
var i, j:Integer;
    s:string;
begin
 try
   parsed.Clear();

   ExtractStrings([#13, #10], [], PChar(text), parsed);

   if (LeftStr(parsed[0], 9) = 'Konec hry') then
    begin
     Application.MessageBox(PChar(parsed[0]), 'Konec hry', MB_OK OR MB_ICONINFORMATION);
     Self.game_id := -1;
     Self.ShowInfo();
     Self.CheckMvBtnEnabled();
     Exit();
    end;

   if (LeftStr(parsed[0], 5) = 'Chyba') then
    begin
     Application.MessageBox(PChar(parsed[0]), 'Chyba', MB_OK OR MB_ICONWARNING);
     Exit();
    end;

   Self.me.base    := Point(-1, -1);
   Self.other.base := Point(-1, -1);

   Self.game_id    := StrToInt(parsed[0]);
   Self.zbyva_tahu := StrToInt(parsed[1]);

   // parsovani mapy
   for i := 0 to _MAP_SIZE-1 do
    begin
     for j := 0 to _MAP_SIZE-1 do
      begin
       s := Copy(parsed[2+i], (j*2)+1, 2);
       if (StrToIntDef(s, -1) <> -1) then
        begin
         Self.map.mapa[j][i].typ := poklad;
         Self.map.mapa[j][i].hodnota_pokladu := StrToInt(s);
        end else begin
         case (s[1]) of
          '+' : begin
                  Self.map.mapa[j][i].typ := prazdne;
                  Self.me.base := Point(j, i);
                end;
          '-' : begin
                  Self.map.mapa[j][i].typ := prazdne;
                  Self.other.base := Point(j, i);
                end;
          '.' : Self.map.mapa[j][i].typ := prazdne;
          'X' : Self.map.mapa[j][i].typ := zed;
         end;//case
        end;
      end;//for j
    end;//for i

   // nas lovec pokladu: S R P C
   Self.parsed2.Clear();
   ExtractStrings([' '], [], PChar(parsed[12]), parsed2);
   Self.me.pos          := Point(StrToInt(parsed2[0])-1, StrToInt(parsed2[1])-1);
   Self.me.nese_hodnotu := StrToInt(parsed2[2]);
   Self.me.prineseno    := StrToInt(parsed2[3]);

   Self.other_enabled := (parsed.Count > 13);
   if (parsed.Count > 13) then
    begin
     Self.parsed2.Clear();
     ExtractStrings([' '], [], PChar(parsed[13]), parsed2);
     Self.other.pos          := Point(StrToInt(parsed2[0])-1, StrToInt(parsed2[1])-1);
     Self.other.nese_hodnotu := StrToInt(parsed2[2]);
     Self.other.prineseno    := StrToInt(parsed2[3]);
    end else begin
     Self.other.pos := Point(-1, -1);
    end;

 except

 end;

 Self.ShowMap();
 Self.ShowInfo();
 Self.CheckMvBtnEnabled();

 if (Self.CHB_AutoUI.Checked) then Self.T_UI.Enabled := true;
end;//procedure

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.ShowInfo();
begin
 Self.L_Nese.Caption       := IntToStr(Self.me.nese_hodnotu);
 Self.L_Odneseno.Caption   := IntToStr(Self.me.prineseno);
 Self.L_Tahu_zbyva.Caption := IntToStr(Self.zbyva_tahu);
end;//procedure

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.SendMove(move:TMove);
var IdHTTP: TIdHTTP;
    mv: char;
begin
 case (move) of
  TMove.doleva  : mv := 'L';
  TMove.doprava : mv := 'P';
  TMove.nahoru  : mv := 'N';
  TMove.dolu    : mv := 'D';
 else
  mv := ' ';
 end;

 IdHTTP := TIdHTTP.Create(nil);
 try
   Self.ParseResponse(IdHTTP.Get('http://'+F_Options.E_hostname.Text+':'+IntToStr(F_Options.SE_Port.Value)+'/tah?id='+IntToStr(Self.game_id)+'&pohyb='+mv));
 finally
   IdHTTP.Free;
 end;
end;//procedure

////////////////////////////////////////////////////////////////////////////////

procedure TF_Main.CHB_KeyboardClick(Sender: TObject);
begin
 Self.CheckMvBtnEnabled();
end;

procedure TF_Main.CheckMvBtnEnabled();
begin
 Self.B_Up.Enabled       := ((Self.me.pos.Y > 0) and (Self.map.mapa[Self.me.pos.X][Self.me.pos.Y-1].typ <> zed)) and (not Self.CHB_Keyboard.Checked) and (Self.game_id > 0);
 Self.B_Left.Enabled     := ((Self.me.pos.X > 0) and (Self.map.mapa[Self.me.pos.X-1][Self.me.pos.Y].typ <> zed)) and (not Self.CHB_Keyboard.Checked) and (Self.game_id > 0);
 Self.B_Down.Enabled     := ((Self.me.pos.Y < _MAP_SIZE-1) and (Self.map.mapa[Self.me.pos.X][Self.me.pos.Y+1].typ <> zed)) and (not Self.CHB_Keyboard.Checked) and (Self.game_id > 0);
 Self.B_Right.Enabled    := ((Self.me.pos.X < _MAP_SIZE-1) and (Self.map.mapa[Self.me.pos.X+1][Self.me.pos.Y].typ <> zed)) and (not Self.CHB_Keyboard.Checked) and (Self.game_id > 0);
 Self.B_UI_Step.Enabled  := (Self.game_id > 0);
 Self.CHB_AutoUI.Enabled := (Self.game_id > 0);
 if (Self.game_id < 0) then Self.CHB_AutoUI.Checked := false; 
end;//procedure

////////////////////////////////////////////////////////////////////////////////
// Pocitacova inteligence:
////////////////////////////////////////////////////////////////////////////////

// Jelikoz hrany nejsou ohodnocene, pouzijeme bezne prohledavani do sirky:
function TF_Main.GetPath(start,fin:TPoint; ignore_poklad:boolean = false):TPath;
var map_ohodnoceno:array[0.._MAP_SIZE-1, 0.._MAP_SIZE-1] of Integer;     // zde si pamatujeme ohodnoceni bodu v bludisti
  i, j: Integer;
  q:TQueue<TPoint>;                                                      // fronta bodu bludiste (vrcholu)
  cur:TPoint;                                                            // aktualne zpracovavany bod
begin
 // smer: nejdriv od fin do start, pak zpetna rekonstrukce ze start do fin -> vrati cestu start->fin

 // nejdriv vsechna ohodnoceni nastavime na -1
 for i := 0 to _MAP_SIZE-1 do
   for j := 0 to _MAP_SIZE-1 do
     map_ohodnoceno[i][j] := -1;
 map_ohodnoceno[fin.X][fin.Y] := 0;             // a ohodnocemi cile na 0 (zaciname od cile)

 // vytvorime frontu a vlozime do ni 0. vrchol (fin)
 q := TQueue<TPoint>.Create();
 q.Enqueue(fin);

 // zpracovavame, dokud jsou ve fronte body
 while (q.Count <> 0) do
  begin
   cur := q.Dequeue();

   // podminka nalezeni cesty
   if ((cur.X = start.X) and (cur.Y = start.Y)) then break;

   // pohyb doprava
   if ((cur.X+1 < _MAP_SIZE) and (map_ohodnoceno[cur.X+1][cur.Y] = -1) and (Self.map.mapa[cur.X+1][cur.Y].typ <> zed) and ((Self.map.mapa[cur.X+1][cur.Y].typ <> poklad) or (ignore_poklad))) then
    begin
     map_ohodnoceno[cur.X+1][cur.Y] := map_ohodnoceno[cur.X][cur.Y]+1;
     q.Enqueue(Point(cur.X+1, cur.Y));
    end;
   // pohyb doleva
   if (cur.X-1 >= 0) and (map_ohodnoceno[cur.X-1][cur.Y] = -1) and (Self.map.mapa[cur.X-1][cur.Y].typ <> zed) and ((Self.map.mapa[cur.X-1][cur.Y].typ <> poklad) or (ignore_poklad)) then
    begin
     map_ohodnoceno[cur.X-1][cur.Y] := map_ohodnoceno[cur.X][cur.Y]+1;
     q.Enqueue(Point(cur.X-1, cur.Y));
    end;
   // pohyb dolu
   if (cur.Y+1 < _MAP_SIZE) and (map_ohodnoceno[cur.X][cur.Y+1] = -1) and (Self.map.mapa[cur.X][cur.Y+1].typ <> zed) and ((Self.map.mapa[cur.X][cur.Y+1].typ <> poklad) or (ignore_poklad)) then
    begin
     map_ohodnoceno[cur.X][cur.Y+1] := map_ohodnoceno[cur.X][cur.Y]+1;
     q.Enqueue(Point(cur.X, cur.Y+1));
    end;
   // pohyb nahoru
   if (cur.Y-1 >= 0) and (map_ohodnoceno[cur.X][cur.Y-1] = -1) and (Self.map.mapa[cur.X][cur.Y-1].typ <> zed) and ((Self.map.mapa[cur.X][cur.Y-1].typ <> poklad) or (ignore_poklad)) then
    begin
     map_ohodnoceno[cur.X][cur.Y-1] := map_ohodnoceno[cur.X][cur.Y]+1;
     q.Enqueue(Point(cur.X, cur.Y-1));
    end;
  end;//while

 q.Free();

 // kontrola nelezeni cesty
 if ((cur.X <> start.X) or (cur.Y <> start.Y)) then
   raise Exception.Create('Cesta nenalezena');

 // zpetna rekonstrukce cesty:
 Result := TList<TMove>.Create();

 while ((cur.X <> fin.X) or (cur.Y <> fin.Y)) do
  begin
   if ((cur.X+1 < _MAP_SIZE) and (map_ohodnoceno[cur.X+1, cur.Y] >= 0) and (map_ohodnoceno[cur.X+1, cur.Y] < map_ohodnoceno[cur.X, cur.Y])) then
    begin
     cur := Point(cur.X+1, cur.Y);
     Result.Add(doprava);
    end;
   if ((cur.X-1 >= 0) and (map_ohodnoceno[cur.X-1, cur.Y] >= 0) and (map_ohodnoceno[cur.X-1, cur.Y] < map_ohodnoceno[cur.X, cur.Y])) then
    begin
     cur := Point(cur.X-1, cur.Y);
     Result.Add(doleva);
    end;
   if ((cur.Y+1 < _MAP_SIZE) and (map_ohodnoceno[cur.X, cur.Y+1] >= 0) and (map_ohodnoceno[cur.X, cur.Y+1] < map_ohodnoceno[cur.X, cur.Y])) then
    begin
     cur := Point(cur.X, cur.Y+1);
     Result.Add(dolu);
    end;
   if ((cur.Y-1 >= 0) and (map_ohodnoceno[cur.X, cur.Y-1] >= 0) and (map_ohodnoceno[cur.X, cur.Y-1] < map_ohodnoceno[cur.X, cur.Y])) then
    begin
     cur := Point(cur.X, cur.Y-1);
     Result.Add(nahoru);
    end;
  end;//while
end;//function

////////////////////////////////////////////////////////////////////////////////

function TF_Main.GetPathLen(start,fin:TPoint; ignore_poklad:boolean = false):Integer;
var res:TList<TMove>;
begin
 res := Self.GetPath(start, fin, ignore_poklad);
 Result := res.Count+1;
 res.Free();
end;//function

////////////////////////////////////////////////////////////////////////////////
// tato funkce vybere takovy poklad, ke kteremu ma cesta nejlepsi pomer vydelek/delka cesty

function TF_Main.VyberNejlepsiPoklad():TPoint;
var x, y:Integer;
    best_pomer:Real;
    best_poklad:TPoint;
    len:Integer;
begin
 best_pomer  := 0;
 best_poklad := Point(-1, -1);

 for x := 0 to _MAP_SIZE do
  begin
   for y := 0 to _MAP_SIZE-1 do
    begin
     if (Self.map.mapa[x][y].typ <> poklad) then continue;

     try
       len := Self.GetPathLen(Self.me.pos, Point(x, y));
     except
       continue;    // klidne se muze stat, ze cesta nebude nalezena -> pokracujeme v hledani dalsiho pokladu
     end;

     if ((Self.map.mapa[x][y].hodnota_pokladu/len) > best_pomer) then
      begin
       best_pomer  := (Self.map.mapa[x][y].hodnota_pokladu/len);
       best_poklad := Point(x, y);
      end;
    end;//for y
  end;//for x

 Result := best_poklad;
end;//function

////////////////////////////////////////////////////////////////////////////////
// provede umelou inteligenci:

procedure TF_Main.DoUI();
var best_poklad:TPoint;
begin
 case (Self.UI.status) of
  TUIStatus.off: begin
   if (not Self.CHB_DeepMode.Checked) then
    begin
     // klasicke hledani
     best_poklad := Self.VyberNejlepsiPoklad();
     if (best_poklad.X < 0) then
      begin
       Exit();
       Self.UI.status := TUIStatus.off;
       Self.CHB_AutoUI.Checked := false;
      end;
     if (Assigned(Self.UI.cmds)) then FreeAndNil(self.UI.cmds);
     Self.UI.cmds := Self.GetPath(Self.me.pos, best_poklad);
    end else begin
     // hledani do hloubky
     if (Assigned(Self.UI.cmds)) then FreeAndNil(self.UI.cmds);
     Self.UI.cmds := Self.GetPath(Self.me.pos, Self.best_solution.poklady[Self.best_solution.poklady.Count-1]);
     Self.best_solution.poklady.Delete(Self.best_solution.poklady.Count-1);
    end;
     Self.UI.status := TUIStatus.mam_cestu_k_pokladu;
     Self.DoUI();
  end;

  TUIStatus.mam_cestu_k_pokladu: begin
   if (self.UI.cmds.Count > 0) then
    begin
     Self.SendMove(Self.UI.cmds[0]);
     Self.UI.cmds.Delete(0);
    end else begin
     // potreba spocitat cestu zpet
     Self.UI.status := TUIStatus.vracim_se_zpet;
     Self.UI.cmds.Free();
     Self.UI.cmds := Self.GetPath(Self.me.pos, Self.me.base);
     Self.DoUI();
    end;
  end;

  TUIStatus.vracim_se_zpet: begin
   if (self.UI.cmds.Count > 0) then
    begin
     Self.SendMove(Self.UI.cmds[0]);
     Self.UI.cmds.Delete(0);
    end else begin
     // vratil jsem se -> pro dalsi poklad
     Self.UI.status := TUIStatus.off;
     Self.DoUI();
    end;
  end;

 end;//case
end;//procedure

////////////////////////////////////////////////////////////////////////////////
// EXPERIMETALNI PROHLEDAVANI ::
////////////////////////////////////////////////////////////////////////////////
// Vybira nejlepsi poklady: uvazuje i to, ze pristupu k velkemu pokladu
//   muze branit maly poklad
// Jak to funguje: spoustime prohledavani vsech moznych variant do budoucna
// Toto probehne docela v rozumnem case, protoze je omezeny pocet kroku.

// Toto prohledavani je EXPERIMENTALNI a zatim nefunguje tak, jak ma.
// Toto prohledavani by melo zvladnout "navnadu", ale zatim se tak bohuzel nedeje.

function TF_Main.VyberDalsiPoklad(aktualni_delka_cesty:Integer):TSolTree;
var x, y, len:Integer;
    best_poklad:TPoint;
    sol:TSolTree;
begin
 best_poklad    := Point(-1, -1);
 Result.zisk    := 0;
 Result.poklady := nil;

 for x := 0 to _MAP_SIZE do
  begin
   for y := 0 to _MAP_SIZE-1 do
    begin
     if (Self.map.mapa[x][y].typ <> poklad) then continue;

     try
       len := Self.GetPathLen(me.base, Point(x, y), false);
     except
       // cesta opravdu neexisutje (poklad je napr. zazdeny -> pokracujeme na dalsi poklad)
       continue;
     end;

     // prekrocili jsme maximalni pocet tahu -> hledame jinou cestu
     if (aktualni_delka_cesty+len > Self.zbyva_tahu) then
      continue;

     Self.map.mapa[x][y].typ := TMapPolickoTyp.prazdne;
     // rekurzivne se zanorime
     sol := Self.VyberDalsiPoklad(aktualni_delka_cesty + len);  // nejlepsi zisk z dane rekurzivni vetve (z celeho podstromu)
     Self.map.mapa[x][y].typ := TMapPolickoTyp.poklad;

     // ktera podvetev mela nejvetsi zisk?
     if ((sol.zisk > Result.zisk) or (Result.zisk = 0)) then
      begin
       if (result.poklady <> nil) then FreeAndNil(result.poklady);
       Result := sol;
       best_poklad := Point(x, y);
     end;
    end;//for y
  end;//for x

 if (not Assigned(Result.poklady)) then
   Result.poklady := TList<TPoint>.Create();
 if (best_poklad.X > -1) then
  begin
   Result.zisk := Result.zisk + Self.map.mapa[best_poklad.X][best_poklad.Y].hodnota_pokladu;
   Result.poklady.Add(best_poklad);
  end;
end;//function

procedure TF_Main.VyberNejlepsiReseni();
begin
 Self.best_solution := Self.VyberDalsiPoklad(0);
end;//procedure

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

end.//unit
